import streamlit as st
import pandas as pd
import numpy as np
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from sklearn.ensemble import RandomForestRegressor
import random
from datetime import datetime, timedelta
import bcrypt
import os

# File paths
USERS_CSV = 'users.csv'
TRAITS_CSV = 'traits.csv'
FEEDBACK_CSV = 'feedback.csv'

# Function to hash passwords
def hash_password(password):
    return bcrypt.hashpw(password.encode(), bcrypt.gensalt()).decode()

# Function to verify password
def verify_password(stored_password, provided_password):
    return bcrypt.checkpw(provided_password.encode(), stored_password.encode())

# Function to register a new user
def register_user(first_name, email, password):
    if os.path.exists(USERS_CSV):
        users_df = pd.read_csv(USERS_CSV)
        if email in users_df['email'].values:
            return False, "Email already registered"
    else:
        users_df = pd.DataFrame(columns=['id', 'first_name', 'email', 'password'])
    
    hashed_password = hash_password(password)
    new_id = len(users_df) + 1
    new_user = pd.DataFrame({'id': [new_id], 'first_name': [first_name], 'email': [email], 'password': [hashed_password]})
    users_df = pd.concat([users_df, new_user], ignore_index=True)
    users_df.to_csv(USERS_CSV, index=False)
    return True, "User registered successfully"

# Function to login a user
def login_user(email, password):
    if not os.path.exists(USERS_CSV):
        return False, "No users found. Please register first."
    
    users_df = pd.read_csv(USERS_CSV)
    user = users_df[users_df['email'] == email]
    if user.empty:
        return False, "User not found"
    
    stored_password = user['password'].values[0]
    if verify_password(stored_password, password):
        return True, user.iloc[0]
    else:
        return False, "Incorrect password"

# Function to save feedback to CSV
def save_feedback(user_id, feedback):
    try:
        if os.path.exists(FEEDBACK_CSV):
            feedback_df = pd.read_csv(FEEDBACK_CSV)
        else:
            feedback_df = pd.DataFrame(columns=['user_id', 'feedback', 'timestamp', 'resolved'])
        
        new_feedback = pd.DataFrame({
            'user_id': [user_id],
            'feedback': [feedback],
            'timestamp': [datetime.now()],
            'resolved': [False]
        })
        feedback_df = pd.concat([feedback_df, new_feedback], ignore_index=True)
        feedback_df.to_csv(FEEDBACK_CSV, index=False)
        st.success("Thank you for your feedback!")
    except Exception as e:
        st.error(f"Error saving feedback: {e}")

# Function to get feedback from CSV
def get_feedback():
    try:
        if os.path.exists(FEEDBACK_CSV):
            feedback_df = pd.read_csv(FEEDBACK_CSV)
            users_df = pd.read_csv(USERS_CSV)
            merged_df = pd.merge(feedback_df, users_df, left_on='user_id', right_on='id')
            return merged_df[['first_name', 'email', 'feedback', 'timestamp', 'resolved']]
        else:
            return pd.DataFrame(columns=['first_name', 'email', 'feedback', 'timestamp', 'resolved'])
    except Exception as e:
        st.error(f"Error reading feedback: {e}")
        return pd.DataFrame()

# Function to save trait measurement to CSV
def save_trait_measurement(user_id, trait1, value1, trait2, value2, outcome):
    try:
        if os.path.exists(TRAITS_CSV):
            traits_df = pd.read_csv(TRAITS_CSV)
        else:
            traits_df = pd.DataFrame(columns=['user_id', 'trait1', 'value1', 'trait2', 'value2', 'outcome', 'timestamp'])
        
        new_trait = pd.DataFrame({
            'user_id': [user_id],
            'trait1': [trait1],
            'value1': [value1],
            'trait2': [trait2],
            'value2': [value2],
            'outcome': [outcome],
            'timestamp': [datetime.now()]
        })
        traits_df = pd.concat([traits_df, new_trait], ignore_index=True)
        traits_df.to_csv(TRAITS_CSV, index=False)
        st.success('Measurement saved successfully!')
    except Exception as e:
        st.error(f"Error saving data: {e}")

# Function to get user traits from CSV
def get_user_traits(user_id):
    try:
        if os.path.exists(TRAITS_CSV):
            traits_df = pd.read_csv(TRAITS_CSV)
            user_traits = traits_df[traits_df['user_id'] == user_id]
            if not user_traits.empty:
                return user_traits
        return generate_fictitious_data(user_id)
    except Exception as e:
        st.error(f"Error reading data: {e}")
        return pd.DataFrame()

# Function to generate fictitious data
def generate_fictitious_data(user_id, num_entries=20):
    data = []
    date = datetime.now() - timedelta(days=num_entries)
    for _ in range(num_entries):
        trait1, trait2 = random.choice(list(opposing_drives.values()))
        value1 = random.uniform(0, 100)
        value2 = random.uniform(0, 100)
        outcome = calculate_effectiveness(value1, value2)
        data.append({
            'user_id': user_id,
            'trait1': trait1,
            'value1': value1,
            'trait2': trait2,
            'value2': value2,
            'outcome': outcome,
            'timestamp': date.strftime("%Y-%m-%d %H:%M:%S")
        })
        date += timedelta(days=1)
    return pd.DataFrame(data)

# Function to calculate effectiveness based on trait values
def calculate_effectiveness(value1, value2):
    balance = 100 - abs(value1 - value2)
    average = (value1 + value2) / 2
    return (balance * 0.5 + average * 0.5)

# Define opposing drives for leaders and individuals
leader_opposing_drives = {
    "Vision vs Execution": ("Vision", "Execution"),
    "Boldness vs Caution": ("Boldness", "Caution"),
    "Empathy vs Decisiveness": ("Empathy", "Decisiveness"),
    "Innovation vs Stability": ("Innovation", "Stability"),
    "Strategic Thinking vs Tactical Action": ("Strategic Thinking", "Tactical Action")
}

individual_opposing_drives = {
    "Creativity vs Discipline": ("Creativity", "Discipline"),
    "Spontaneity vs Consistency": ("Spontaneity", "Consistency"),
    "Introversion vs Extroversion": ("Introversion", "Extroversion")
}

# Trait descriptions
trait_descriptions = {
    "Vision": "The ability to see the big picture and set long-term goals.",
    "Execution": "The ability to implement plans and achieve short-term goals.",
    "Boldness": "The willingness to take risks and make decisive moves.",
    "Caution": "The tendency to be careful and avoid unnecessary risks.",
    "Empathy": "The ability to understand and share the feelings of others.",
    "Decisiveness": "The ability to make quick and firm decisions.",
    "Innovation": "The drive to create and implement new ideas and solutions.",
    "Stability": "The focus on maintaining consistency and reliability.",
    "Strategic Thinking": "The ability to plan for the long-term and consider the broader implications of decisions.",
    "Tactical Action": "The focus on immediate tasks and short-term goals.",
    "Creativity": "The ability to generate new and original ideas.",
    "Discipline": "The ability to maintain focus and follow through on tasks.",
    "Spontaneity": "The ability to act on impulse and embrace the unexpected.",
    "Consistency": "The ability to maintain a steady and predictable approach.",
    "Introversion": "The tendency to be reflective and enjoy solitary activities.",
    "Extroversion": "The tendency to be outgoing and enjoy social interactions."
}

# Initialize session state for login
if 'logged_in' not in st.session_state:
    st.session_state['logged_in'] = False
if 'menu' not in st.session_state:
    st.session_state['menu'] = "Login"

# Sidebar for navigation
st.sidebar.title("Navigation")
menu = st.sidebar.radio("Go to", ["Login", "Application", "Feedback"], index=["Login", "Application", "Feedback"].index(st.session_state['menu']))

# Logout button
if st.session_state['logged_in']:
    if st.sidebar.button("Logout"):
        st.session_state['logged_in'] = False
        st.session_state['menu'] = "Login"
        st.experimental_rerun()

# Login page
if menu == "Login":
    st.title("Login / Register")
    tab1, tab2 = st.tabs(["Login", "Register"])
    with tab1:
        with st.form(key='login_form'):
            email = st.text_input("Email")
            password = st.text_input("Password", type="password")
            login_button = st.form_submit_button(label='Login')
            if login_button:
                if email and password:
                    success, result = login_user(email, password)
                    if success:
                        st.session_state['logged_in'] = True
                        st.session_state['user_id'] = result['id']
                        st.session_state['first_name'] = result['first_name']
                        st.session_state['email'] = result['email']
                        st.session_state['menu'] = "Application"
                        st.success("Logged in successfully!")
                        st.experimental_rerun()
                    else:
                        st.error(result)
                else:
                    st.error("Please enter both email and password.")
    with tab2:
        with st.form(key='register_form'):
            first_name = st.text_input("First Name")
            email = st.text_input("Email")
            password = st.text_input("Password", type="password")
            consent = st.checkbox("I consent to the processing of my personal data in accordance with the privacy policy.")
            register_button = st.form_submit_button(label='Register')
            if register_button:
                if first_name and email and password and consent:
                    success, message = register_user(first_name, email, password)
                    if success:
                        st.success(message)
                    else:
                        st.error(message)
                else:
                    st.error("Please fill in all fields and provide consent.")

# Application page
if menu == "Application" and st.session_state['logged_in']:
    st.title("TheParadoxicalTracker")
    st.subheader(f"Welcome, {st.session_state['first_name']}!")

    # User input
    user_id = st.session_state['user_id']

    # User type selection
    user_type = st.selectbox("Are you a leader or an individual?", ["Leader", "Individual"])

    # Select opposing drives based on user type
    opposing_drives = leader_opposing_drives if user_type == "Leader" else individual_opposing_drives

    # Let user select which opposing drive to focus on
    selected_drive = st.selectbox("Select an opposing drive to focus on:", list(opposing_drives.keys()))
    st.subheader(f"Rate Your Traits: {selected_drive}")
    trait1, trait2 = opposing_drives[selected_drive]

    col1, col2 = st.columns(2)
    with col1:
        value1 = st.slider(f'{trait1}', 0.0, 100.0, 50.0, key='slider1')
    with col2:
        value2 = st.slider(f'{trait2}', 0.0, 100.0, 50.0, key='slider2')

    # Calculate initial effectiveness
    calculated_effectiveness = calculate_effectiveness(value1, value2)

    # Allow for manual adjustment of effectiveness
    st.write("Overall Effectiveness (auto-calculated, adjust if needed):")
    outcome = st.slider('', 0.0, 100.0, float(calculated_effectiveness), key='effectiveness_slider')

    if st.button('Save Measurement'):
        save_trait_measurement(user_id, trait1, value1, trait2, value2, outcome)

    # Display user's traits
    user_traits = get_user_traits(user_id)
    if not user_traits.empty:
        st.subheader(f'Trait History for User {user_id}')
        st.write(user_traits)

    # Create the interactive dashboard
    fig = make_subplots(
        rows=2, cols=2,
        subplot_titles=("Current Trait Levels", "Trait Pattern Over Time", "Predicted Outcome", "Feature Importance"),
        specs=[[{"type": "bar", "colspan": 1}, {"type": "scatter", "colspan": 1}],
               [{"type": "indicator", "colspan": 1}, {"type": "bar", "colspan": 1}]],
        column_widths=[0.3, 0.7],
        row_heights=[0.5, 0.5]
    )

    # Bar chart for current trait levels
    current_traits = pd.DataFrame({
        'Trait': [trait1, trait2],
        'Value': [value1, value2]
    })
    fig.add_trace(
        go.Bar(x=current_traits['Trait'], y=current_traits['Value'], name="Current Traits", marker_color='#e8491d'),
        row=1, col=1
    )

    # Area chart for trait pattern over time
    fig.add_trace(
        go.Scatter(x=user_traits['timestamp'], y=user_traits['value1'], fill='tozeroy', name=trait1, line=dict(color='#e8491d')),
        row=1, col=2
    )
    fig.add_trace(
        go.Scatter(x=user_traits['timestamp'], y=user_traits['value2'], fill='tozeroy', name=trait2, line=dict(color='#35424a')),
        row=1, col=2
    )

    # Prediction model
    if len(user_traits) >= 10:
        X = user_traits[['value1', 'value2']]
        y = user_traits['outcome']
        model = RandomForestRegressor(n_estimators=100, random_state=42)
        model.fit(X, y)
        predicted_outcome = model.predict([[value1, value2]])[0]

        # Indicator for predicted outcome
        fig.add_trace(
            go.Indicator(
                mode="gauge+number",
                value=predicted_outcome,
                title={'text': "Predicted Outcome"},
                gauge={'axis': {'range': [0, 100]}, 'bar': {'color': '#e8491d'}},
            ),
            row=2, col=1
        )

        # Feature importance
        importances = pd.DataFrame({
            'Trait': [trait1, trait2],
            'Importance': model.feature_importances_
        })
        fig.add_trace(
            go.Bar(x=importances['Trait'], y=importances['Importance'], name="Feature Importance", marker_color='#35424a'),
            row=2, col=2
        )
    else:
        # Add empty traces with annotations when there's not enough data
        fig.add_trace(go.Indicator(mode="number", value=0), row=2, col=1)
        fig.add_trace(go.Bar(x=[], y=[]), row=2, col=2)
        
        fig.add_annotation(
            text="Not enough data for predictions",
            showarrow=False,
            xref="paper", yref="paper",
            x=0.25, y=0.25,
            xanchor="center", yanchor="middle"
        )
        fig.add_annotation(
            text="Not enough data for feature importance",
            showarrow=False,
            xref="paper", yref="paper",
            x=0.75, y=0.25,
            xanchor="center", yanchor="middle"
        )

    # Update layout
    fig.update_layout(
        height=800,
        width=1200,
        title_text="Trait Dashboard",
        plot_bgcolor='rgba(0,0,0,0)',
        paper_bgcolor='rgba(0,0,0,0)',
        font=dict(color='#2c3e50')
    )

    # Display the figure in Streamlit
    st.plotly_chart(fig)

    # Calculate and display balance
    balance = np.abs(value1 - value2)
    st.write(f'{trait1} vs {trait2} Balance: {balance:.2f}')
    if balance < 20:
        st.write(f"Your {trait1} and {trait2} are well-balanced.")
    elif value1 > value2:
        st.write(f"You're leaning more towards {trait1}. Consider embracing more {trait2} in certain situations.")
    else:
        st.write(f"You're leaning more towards {trait2}. There might be situations where more {trait1} is beneficial.")

    st.subheader("Embracing Paradoxical Traits")
    st.write(f"""
    Effective leadership often requires balancing seemingly opposing traits. In this case, we're focusing on {trait1} vs {trait2}:
    - {trait1}: {trait_descriptions[trait1]}
    - {trait2}: {trait_descriptions[trait2]}
    
    The key is to cultivate both these traits and apply them contextually:
    - Recognize which trait is most appropriate for each situation.
    - Strive for a dynamic balance rather than a static middle ground.
    - Reflect on your style and consciously develop your less dominant trait.
    - Remember, true paradox is not about compromise, but about embracing the tension between opposing drives.
    
    Consider situations where you might need to lean more towards {trait1}, and others where {trait2} might be more beneficial. The goal is to develop the ability to flex between these traits as the situation demands.
    """)

# Feedback page
if menu == "Feedback" and st.session_state['logged_in']:
    st.title("Feedback")
    st.subheader("Leave Your Feedback")
    with st.form(key='feedback_form'):
        feedback = st.text_area("Feedback")
        submit_button = st.form_submit_button(label='Submit')
        if submit_button:
            save_feedback(st.session_state['user_id'], feedback)
    # Display feedback
    st.subheader("User Feedback")
    feedback_data = get_feedback()
    if not feedback_data.empty:
        st.dataframe(feedback_data[['first_name', 'email', 'feedback', 'timestamp', 'resolved']])

        # Allow download only if the logged-in user is Joyceridge Mensah
        if st.session_state['email'] == "joyceridgem@gmail.com":  # Replace with the actual admin email
            csv = feedback_data.to_csv(index=False)
            st.download_button(
                label="Download Feedback",
                data=csv,
                file_name="feedback.csv",
                mime="text/csv",
            )
    else:
        st.write("No feedback available.")

# Add this at the end of your script to ensure all pages are properly closed
if menu not in ["Login", "Application", "Feedback"]:
    st.error("Invalid menu selection.")